@@ -6,13 +6,13 @@ Simple one-line integration with zero overhead.
 
 Usage:
     import lace
-    
+
     # Create attestation before training
     attestation_id = lace.attest("./training_data")
-    
+
     # Monitor training (one line)
     lace.monitor()
-    
+
     # Verify after training
     result = lace.verify(attestation_id)
 
@@ -22,26 +22,15 @@ Get your API key at https://withlace.ai
 
 __version__ = "0.7.0rc1"
 
-from .client import (
-    LaceClient,
-    attest,
-    verify,
-    monitor,
-    get_client
-)
+from .client import LaceClient, attest, get_client, monitor, verify
+
+__all__ = ["LaceClient", "attest", "verify", "monitor", "get_client", "__version__"]
 
-__all__ = [
-    'LaceClient',
-    'attest',
-    'verify', 
-    'monitor',
-    'get_client',
-    '__version__'
-]
 
 def about():
     """Display information about Lace."""
-    print(f"""
+    print(
+        f"""
 Lace v{__version__} - AI Training Transparency Protocol
 ========================================================
 Prevent copyright lawsuits with cryptographic proof of training.
@@ -52,4 +41,5 @@ Quick Start:
 
 Learn more: https://withlace.ai
 Get API key: https://withlace.ai/request-demo
-    """)
+    """
+    )
