@@ -4,4 +4,4 @@ Regulatory compliance modules for EU AI Act.
 
 from .scope import ScopeClassifier, ScopeResult
 
-__all__ = ['ScopeClassifier', 'ScopeResult']
\ No newline at end of file
+__all__ = ["ScopeClassifier", "ScopeResult"]
